package PObject;
# LICENSE
#
# Copyright (c) 2001 Steve Slaven. All rights reserved.
#
# This program is free software  and is provided "as is" without
# express or implied warranty. You can redistribute it and/or
# modify it under the same terms as Perl itself.
#

# This is a good base class for other objects, includes auto-binding
# for attributes and stuff.

use Carp;
use strict;

my %PackageDebugs = ();

sub new
{
    my $Check = shift;
    my $Class = ref( $Check ) || $Check;

    my %AutoSet;

    my( $HashRef, $ReadRef, $Self, @ThisISA, $UseAuto );
    my $FindPack;

    # Bail if we're a POBJ
    if( $Class eq "PObject" )
    {
        return( {} );
    }

    # This creates a ptr to the original fields
    # to check if they're OK, and inserts the defaults
    # in to this object.  Using a symbolic ref so can't have strict refs
    $FindPack = $Class;
    no strict 'refs';
    @ThisISA = eval( '@' . $FindPack . "::ISA" );
    $HashRef = eval( '\%' . $FindPack . "::AUTOFIELDS" );
    $ReadRef = eval( '\%' . $FindPack . "::READFIELDS" );
    $UseAuto = eval( '$' . $FindPack . "::AUTOSETTINGS" );
    use strict 'refs';

    if( ! $HashRef )
    {
	$HashRef = {};
    }

    if( ! $ReadRef )
    {
	$ReadRef = {};
    }

    # First check through the ISA ref to find any parents.  Call new on them.
    if( @ThisISA )
    {
        $Self = ( shift( @ThisISA ) -> new() );
    }
    else
    {
        $Self = {};
    }

    bless $Self, $Class;

    # Add in the things
    $Self -> { _POBJ_DEBUG } = 0;

    for ( keys( %{ $HashRef } ) )
    {
	$Self -> _dmesg( "Adding autofield: $_", 3 );
        $Self -> { $_ } = $HashRef -> { $_ };
	$Self -> { _POBJ_AUTOFIELD } -> { $_ } = 1;
    }

    for ( keys( %{ $ReadRef } ) )
    {
	$Self -> _dmesg( "Adding readfield: $_", 3 );
        $Self -> { $_ } = $ReadRef -> { $_ };
	$Self -> { _POBJ_READFIELD } -> { $_ } = 1;
    }
   
    # Do the auto-settings
    if( $UseAuto )
    {
        $Self = $Self -> OBJ_NEW();

	if( $Self )
	{
		$Self -> _dmesg( "Using AUTOSETTINGS", 3 );
		%AutoSet = @_;
        	for ( keys( %AutoSet ) )
        	{
	    		$Self -> _dmesg( $_ . ":" . $AutoSet{ $_ }, 3 );
            		$Self -> $_( $AutoSet{ $_ } );
        	}
	}
    }
    else
    {
        $Self = $Self -> OBJ_NEW( @_ );
    }

    return( $Self );
}

sub AUTOLOAD
{
    my $Self = shift;
    my $Arg = shift;
    
    my $Field = $PObject::AUTOLOAD;
    
    $Field =~ s/.*://;
    my $HookMethod;
    
    if( ! ref( $Self ) )
    {
	confess( "Called from package name: " . $Self . " -> " . $Field );
    }

    $Self -> _dmesg( "AUTOLOAD is: '" . $Field . "'", 9 );
    $Self -> _dmesg( "Arg ref is: '" . ref( $Arg ) . "'", 9 );

    $Field = uc( $Field );

    # Check for HOOK
    if( $Field =~ /^OBJ_HOOK/ )
    {
	$Self -> _dmesg( "Failed hook, returning value", 9 );
	return( $Arg );
    }
    
    $Self -> _dmesg( "Checking for AUTOFIELD: " . $Field, 9 );
    if( $Self -> { _POBJ_AUTOFIELD } -> { $Field } )
    {
	# OK do it
	$Self -> _dmesg( "Found AUTOFIELD: " . $Field . "='" . $Self -> { $Field } . "'", 9 );

	if( defined( $Arg ) )
	{
	    $HookMethod = "OBJ_HOOK_" . $Field;
	    $Arg = $Self -> $HookMethod( $Arg );
	
	    if( defined( $Arg ) )
	    {
		$Self -> { $Field } = $Arg;
		$Self -> _dmesg( "Set value: '" . $Self -> { $Field } . "'", 9 );
	    }
	}
	
	return( $Self -> { $Field } );
    }
    else
    {
	# Check for read-only field
	if( $Self -> { _POBJ_READFIELD } -> { $Field } )
	{
	    return( $Self -> { $Field } );
	}
	else
	{
	    # Try to see if they provide a XAD_AUTOLOAD
	    $Self -> _dmesg( "Trying OBJ_AUTOLOAD", 9 );
	    return( $Self -> OBJ_AUTOLOAD( $Field, @_ ) );
	}
    }
}

sub DESTROY
{
    # This needs to be defined because of the AUTOLOAD thing.
}

sub OBJ_AUTOLOAD
{
    # Bad call freak.
    my $Self = shift;
    confess( "Uneditable or unknown attribute called: " . shift() );
}

sub OBJ_NEW
{
    return( shift() );
}


# Debugging functions
sub _debug
{
    # Sets the debugging level
    my $Self = shift;
    my $Debug = shift;

    my $ClassName = ref( $Self ) || $Self;

    if( ref( $Self ) )
    {
	$Self -> { _POBJ_DEBUG } = $Debug;
	$Self -> _dmesg( "DEBUG level set for object", 1 );
    }
    else
    {
	$PackageDebugs{ $ClassName } = $Debug;
	$Self -> _dmesg( "DEBUG level set for package", 1 );
    }
	
    $Self -> _dmesg( "Debug level set to $Debug", 1 );
}

sub _dmesg
{
    my $Self = shift;
    my $Message = shift;
    my $Level = shift;

    my( $Build, $ClassName );

    $ClassName = ref( $Self ) || $Self;
    $Build = ( ref( $Self ) || $Self ) . "[" . $Level . "]) " . $Message;

    if( $PackageDebugs{ $ClassName } >= $Level )
    {
	# Spew regardless
	carp( $Build );
    }
    else
    {
	# Check object level
	if( ref( $Self ) && ( $Self -> { _POBJ_DEBUG } >= $Level ) )
	{
	    carp( $Build );
	}
    }
}

sub _dfatal
{
    # This is just a croak type call
    my $Self = shift;
    my $Message = shift;
    my $Perspective = shift;

    $Self -> _dmesg( $Message, -1, $Perspective );
    croak( "Fatal error..." );
}

sub ISA
{
    my $Self = shift;
    
    return( ref( $Self ) )
}

1;
